# -*- coding: utf-8 -*-

"""
The toolbox is organized around two classes hierarchies: the functions and the
solvers. Instantiated functions represent convex functions to optimize.
Instantiated solvers represent solving algorithms. The
:func:`pyunlocbox.solvers.solve` solving function takes as parameters a solver
object and some function objects to actually solve the optimization problem.
See this function's documentation for a typical usage example.

The :mod:`pyunlocbox` package is divided into the following modules :

* :mod:`pyunlocbox.solvers`: problem solvers, implement the solvers class
  hierarchy and the solving function
* :mod:`pyunlocbox.functions`: functions to be passed to the solvers, implement
  the functions class hierarchy
"""

# When importing the toolbox, you surely want these modules.
from pyunlocbox import functions
from pyunlocbox import solvers
from pyunlocbox import operators

# Silence the code checker warning about unused symbols.
assert functions
assert solvers
assert operators

__version__ = '0.4.0'
__release_date__ = '2016-08-01'



# -*- coding: utf-8 -*-

r"""
This module implements solver objects who minimize an objective function. Call
:func:`solve` to solve your convex optimization problem using your instantiated
solver and functions objects. The :class:`solver` base class defines the
interface of all solver objects. The specialized solver objects inherit from
it and implement the class methods. The following solvers are included :

* :class:`forward_backward`: Forward-backward proximal splitting algorithm.
* :class:`douglas_rachford`: Douglas-Rachford proximal splitting algorithm.
* :class:`generalized_forward_backward`: Generalized Forward-Backward.

* :class:`primal_dual`: Primal-dual algorithms.

  * :class:`mlfbf`: Monotone+Lipschitz Forward-Backward-Forward primal-dual
    algorithm.
  * :class:`projection_based`: Projection-based primal-dual algorithm.
"""

import numpy as np
import time
from pyunlocbox.functions import dummy


def solve(functions, x0, solver=None, atol=None, dtol=None, rtol=1e-3,
          xtol=None, maxit=200, verbosity='LOW'):
    r"""
    Solve an optimization problem whose objective function is the sum of some
    convex functions.

    This function minimizes the objective function :math:`f(x) =
    \sum\limits_{k=0}^{k=K} f_k(x)`, i.e. solves
    :math:`\operatorname{arg\,min}\limits_x f(x)` for :math:`x \in
    \mathbb{R}^{n \times N}` where :math:`n` is the dimensionality of the data
    and :math:`N` the number of independent problems. It returns a dictionary
    with the found solution and some informations about the algorithm
    execution.

    Parameters
    ----------
    functions : list of objects
        A list of convex functions to minimize. These are objects who must
        implement the :meth:`pyunlocbox.functions.func.eval` method. The
        :meth:`pyunlocbox.functions.func.grad` and / or
        :meth:`pyunlocbox.functions.func.prox` methods are required by some
        solvers. Note also that some solvers can only handle two convex
        functions while others may handle more. Please refer to the
        documentation of the considered solver.
    x0 : array_like
        Starting point of the algorithm, :math:`x_0 \in \mathbb{R}^{n \times
        N}`. Note that if you pass a numpy array it will be modified in place
        during execution to save memory. It will then contain the solution. Be
        careful to pass data of the type (int, float32, float64) you want your
        computations to use.
    solver : solver class instance, optional
        The solver algorithm. It is an object who must inherit from
        :class:`pyunlocbox.solvers.solver` and implement the :meth:`_pre`,
        :meth:`_algo` and :meth:`_post` methods. If no solver object are
        provided, a standard one will be chosen given the number of convex
        function objects and their implemented methods.
    atol : float, optional
        The absolute tolerance stopping criterion. The algorithm stops when
        :math:`f(x^t) < atol` where :math:`f(x^t)` is the objective function at
        iteration :math:`t`. Default is None.
    dtol : float, optional
        Stop when the objective function is stable enough, i.e. when
        :math:`\left|f(x^t) - f(x^{t-1})\right| < dtol`. Default is None.
    rtol : float, optional
        The relative tolerance stopping criterion. The algorithm stops when
        :math:`\left|\frac{ f(x^t) - f(x^{t-1}) }{ f(x^t) }\right| < rtol`.
        Default is :math:`10^{-3}`.
    xtol : float, optional
        Stop when the variable is stable enough, i.e. when :math:`\frac{\|x^t -
        x^{t-1}\|_2}{\sqrt{n N}} < xtol`. Note that additional memory will be
        used to store :math:`x^{t-1}`. Default is None.
    maxit : int, optional
        The maximum number of iterations. Default is 200.
    verbosity : {'NONE', 'LOW', 'HIGH', 'ALL'}, optional
        The log level : ``'NONE'`` for no log, ``'LOW'`` for resume at
        convergence, ``'HIGH'`` for info at all solving steps, ``'ALL'`` for
        all possible outputs, including at each steps of the proximal operators
        computation. Default is ``'LOW'``.

    Returns
    -------
    sol : ndarray
        The problem solution.
    solver : str
        The used solver.
    crit : {'ATOL', 'DTOL', 'RTOL', 'XTOL', 'MAXIT'}
        The used stopping criterion. See above for definitions.
    niter : int
        The number of iterations.
    time : float
        The execution time in seconds.
    objective : ndarray
        The successive evaluations of the objective function at each iteration.

    Examples
    --------
    >>> import pyunlocbox
    >>> import numpy as np

    Define a problem:

    >>> y = [4, 5, 6, 7]
    >>> f = pyunlocbox.functions.norm_l2(y=y)

    Solve it:

    >>> x0 = np.zeros(len(y))
    >>> ret = pyunlocbox.solvers.solve([f], x0, atol=1e-2, verbosity='ALL')
    INFO: Dummy objective function added.
    INFO: Selected solver: forward_backward
        norm_l2 evaluation: 1.260000e+02
        dummy evaluation: 0.000000e+00
    INFO: Forward-backward method: FISTA
    Iteration 1 of forward_backward:
        norm_l2 evaluation: 1.400000e+01
        dummy evaluation: 0.000000e+00
        objective = 1.40e+01
    Iteration 2 of forward_backward:
        norm_l2 evaluation: 1.555556e+00
        dummy evaluation: 0.000000e+00
        objective = 1.56e+00
    Iteration 3 of forward_backward:
        norm_l2 evaluation: 3.293044e-02
        dummy evaluation: 0.000000e+00
        objective = 3.29e-02
    Iteration 4 of forward_backward:
        norm_l2 evaluation: 8.780588e-03
        dummy evaluation: 0.000000e+00
        objective = 8.78e-03
    Solution found after 4 iterations:
        objective function f(sol) = 8.780588e-03
        stopping criterion: ATOL

    Verify the stopping criterion (should be smaller than atol=1e-2):

    >>> np.linalg.norm(ret['sol'] - y)**2  # doctest:+ELLIPSIS
    0.00878058...

    Show the solution (should be close to y w.r.t. the L2-norm measure):

    >>> ret['sol']
    array([ 4.03339154,  5.04173943,  6.05008732,  7.0584352 ])

    Show the used solver:

    >>> ret['solver']
    'forward_backward'

    Show some information about the convergence:

    >>> ret['crit']
    'ATOL'
    >>> ret['niter']
    4
    >>> ret['time']  # doctest:+SKIP
    0.0012578964233398438
    >>> ret['objective']  # doctest:+NORMALIZE_WHITESPACE,+ELLIPSIS
    [[126.0, 0], [13.99999999..., 0], [1.55555555..., 0],
    [0.03293043..., 0], [0.00878058..., 0]]

    """

    if verbosity not in ['NONE', 'LOW', 'HIGH', 'ALL']:
        raise ValueError('Verbosity should be either NONE, LOW, HIGH or ALL.')

    # Add a second dummy convex function if only one function is provided.
    if len(functions) < 1:
        raise ValueError('At least 1 convex function should be provided.')
    elif len(functions) == 1:
        functions.append(dummy())
        if verbosity in ['LOW', 'HIGH', 'ALL']:
            print('INFO: Dummy objective function added.')

    # Choose a solver if none provided.
    if not solver:
        if len(functions) == 2:
            fb0 = 'GRAD' in functions[0].cap(x0) and \
                  'PROX' in functions[1].cap(x0)
            fb1 = 'GRAD' in functions[1].cap(x0) and \
                  'PROX' in functions[0].cap(x0)
            dg0 = 'PROX' in functions[0].cap(x0) and \
                  'PROX' in functions[1].cap(x0)
            if fb0 or fb1:
                solver = forward_backward()  # Need one prox and 1 grad.
            elif dg0:
                solver = douglas_rachford()  # Need two prox.
            else:
                raise ValueError('No suitable solver for the given functions.')
        elif len(functions) > 2:
            solver = generalized_forward_backward()
        if verbosity in ['LOW', 'HIGH', 'ALL']:
            name = solver.__class__.__name__
            print('INFO: Selected solver: {}'.format(name))

    # Set solver and functions verbosity.
    translation = {'ALL': 'HIGH', 'HIGH': 'HIGH', 'LOW': 'LOW', 'NONE': 'NONE'}
    solver.verbosity = translation[verbosity]
    translation = {'ALL': 'HIGH', 'HIGH': 'LOW', 'LOW': 'NONE', 'NONE': 'NONE'}
    functions_verbosity = []
    for f in functions:
        functions_verbosity.append(f.verbosity)
        f.verbosity = translation[verbosity]

    tstart = time.time()
    crit = None
    niter = 0
    objective = [[f.eval(x0) for f in functions]]
    rtol_only_zeros = True

    # Solver specific initialization.
    solver.pre(functions, x0)

    while not crit:

        niter += 1

        if xtol is not None:
            last_sol = np.array(solver.sol, copy=True)

        if verbosity in ['HIGH', 'ALL']:
            name = solver.__class__.__name__
            print('Iteration {} of {}:'.format(niter, name))

        # Solver iterative algorithm.
        solver.algo(objective, niter)

        objective.append([f.eval(solver.sol) for f in functions])
        current = np.sum(objective[-1])
        last = np.sum(objective[-2])

        # Verify stopping criteria.
        if atol is not None and current < atol:
            crit = 'ATOL'
        if dtol is not None and np.abs(current - last) < dtol:
            crit = 'DTOL'
        if rtol is not None:
            div = current  # Prevent division by 0.
            if div == 0:
                if verbosity in ['LOW', 'HIGH', 'ALL']:
                    print('WARNING: objective function is equal to 0 !')
                if last != 0:
                    div = last
                else:
                    div = 1.0  # Result will be zero anyway.
            else:
                rtol_only_zeros = False
            relative = np.abs((current - last) / div)
            if relative < rtol and not rtol_only_zeros:
                crit = 'RTOL'
        if xtol is not None:
            err = np.linalg.norm(solver.sol - last_sol)
            err /= np.sqrt(last_sol.size)
            if err < xtol:
                crit = 'XTOL'
        if maxit is not None and niter >= maxit:
            crit = 'MAXIT'

        if verbosity in ['HIGH', 'ALL']:
            print('    objective = {:.2e}'.format(current))

    # Restore verbosity for functions. In case they are called outside solve().
    for k, f in enumerate(functions):
        f.verbosity = functions_verbosity[k]

    if verbosity in ['LOW', 'HIGH', 'ALL']:
        print('Solution found after {} iterations:'.format(niter))
        print('    objective function f(sol) = {:e}'.format(current))
        print('    stopping criterion: {}'.format(crit))

    # Returned dictionary.
    result = {'sol':       solver.sol,
              'solver':    solver.__class__.__name__,  # algo for consistency ?
              'crit':      crit,
              'niter':     niter,
              'time':      time.time() - tstart,
              'objective': objective}
    try:
        result['dual_sol'] = solver.dual_sol
    except AttributeError:
        pass

    # Solver specific post-processing (e.g. delete references).
    solver.post()

    return result


def _prox_star(func, z, T):
    r"""Proximity operator of the convex conjugate of a function."""
    return z - T * func.prox(z / T, 1 / T)


class solver(object):
    r"""
    Defines the solver object interface.

    This class defines the interface of a solver object intended to be passed
    to the :func:`pyunlocbox.solvers.solve` solving function. It is intended to
    be a base class for standard solvers which will implement the required
    methods. It can also be instantiated by user code and dynamically modified
    for rapid testing. This class also defines the generic attributes of all
    solver objects.

    Parameters
    ----------
    step : float
        The gradient-descent step-size. This parameter is bounded by 0 and
        :math:`\frac{2}{\beta}` where :math:`\beta` is the Lipschitz constant
        of the gradient of the smooth function (or a sum of smooth functions).
        Default is 1.
    post_step : function
        User defined function to post-process the step size. This function is
        called every iteration and permits the user to alter the solver
        algorithm. The user may start with a high step size and progressively
        lower it while the algorithm runs to accelerate the convergence. The
        function parameters are the following : `step` (current step size),
        `sol` (current problem solution), `objective` (list of successive
        evaluations of the objective function), `niter` (current iteration
        number). The function should return a new value for `step`. Default is
        to return an unchanged value.
    post_sol : function
        User defined function to post-process the problem solution. This
        function is called every iteration and permits the user to alter the
        solver algorithm. Same parameter as :func:`post_step`. Default is to
        return an unchanged value.
    """

    def __init__(self, step=1, post_step=None, post_sol=None):
        if step < 0:
            raise ValueError('Gamma should be a positive number.')
        self.step = step
        if post_step:
            self.post_step = post_step
        else:
            self.post_step = lambda step, sol, objective, niter: step
        if post_sol:
            self.post_sol = post_sol
        else:
            self.post_sol = lambda step, sol, objective, niter: sol

    def pre(self, functions, x0):
        """
        Solver specific initialization. See parameters documentation in
        :func:`pyunlocbox.solvers.solve` documentation.
        """
        self.sol = np.asarray(x0)
        self._pre(functions, np.asarray(x0))

    def _pre(self, functions, x0):
        raise NotImplementedError("Class user should define this method.")

    def algo(self, objective, niter):
        """
        Call the solver iterative algorithm while allowing the user to alter
        it. This makes it possible to dynamically change the `step` step size
        while the algorithm is running.  See parameters documentation in
        :func:`pyunlocbox.solvers.solve` documentation.
        """
        self._algo()
        self.step = self.post_step(self.step, self.sol, objective, niter)
        self.sol = self.post_sol(self.step, self.sol, objective, niter)

    def _algo(self):
        raise NotImplementedError("Class user should define this method.")

    def post(self):
        """
        Solver specific post-processing. Mainly used to delete references added
        during initialization so that the garbage collector can free the
        memory. See parameters documentation in
        :func:`pyunlocbox.solvers.solve` documentation.
        """
        self._post()
        del self.sol

    def _post(self):
        raise NotImplementedError("Class user should define this method.")


class forward_backward(solver):
    r"""
    Forward-backward proximal splitting algorithm.

    This algorithm solves convex optimization problems composed of the sum of
    a smooth and a non-smooth function.

    See generic attributes descriptions of the
    :class:`pyunlocbox.solvers.solver` base class.

    Parameters
    ----------
    method : {'FISTA', 'ISTA'}, optional
        The method used to solve the problem. It can be 'FISTA' or 'ISTA'. Note
        that while FISTA is much more time efficient, it is less memory
        efficient.  Default is 'FISTA'.
    lambda_ : float, optional
        The update term weight for ISTA. It should be between 0 and 1. Default
        is 1.

    Notes
    -----
    This algorithm requires one function to implement the
    :meth:`pyunlocbox.functions.func.prox` method and the other one to
    implement the :meth:`pyunlocbox.functions.func.grad` method.

    See :cite:`beck2009FISTA` for details about the algorithm.

    Examples
    --------
    >>> from pyunlocbox import functions, solvers
    >>> import numpy as np
    >>> y = [4, 5, 6, 7]
    >>> x0 = np.zeros(len(y))
    >>> f1 = functions.norm_l2(y=y)
    >>> f2 = functions.dummy()
    >>> solver = solvers.forward_backward(method='FISTA', lambda_=1, step=0.5)
    >>> ret = solvers.solve([f1, f2], x0, solver, atol=1e-5)
    Solution found after 12 iterations:
        objective function f(sol) = 4.135992e-06
        stopping criterion: ATOL
    >>> ret['sol']
    array([ 3.99927529,  4.99909411,  5.99891293,  6.99873176])

    """

    def __init__(self, method='FISTA', lambda_=1, *args, **kwargs):
        super(forward_backward, self).__init__(*args, **kwargs)
        self.method = method
        self.lambda_ = lambda_

    def _pre(self, functions, x0):

        if self.verbosity is 'HIGH':
            print('INFO: Forward-backward method: {}'.format(self.method))

        if self.lambda_ <= 0 or self.lambda_ > 1:
            raise ValueError('Lambda is bounded by 0 and 1.')

        if self.method is 'ISTA':
            self._algo = self._ista
        elif self.method is 'FISTA':
            self._algo = self._fista
            self.z = np.array(x0, copy=True)
            self.t = 1.
        else:
            raise ValueError('The method should be either FISTA or ISTA.')

        if len(functions) != 2:
            raise ValueError('Forward-backward requires two convex functions.')

        if 'PROX' in functions[0].cap(x0) and 'GRAD' in functions[1].cap(x0):
            self.f1 = functions[0]
            self.f2 = functions[1]
        elif 'PROX' in functions[1].cap(x0) and 'GRAD' in functions[0].cap(x0):
            self.f1 = functions[1]
            self.f2 = functions[0]
        else:
            raise ValueError('Forward-backward requires a function to '
                             'implement prox() and the other grad().')

    def _ista(self):
        x = self.sol - self.step * self.f2.grad(self.sol)
        self.sol[:] += self.lambda_ * (self.f1.prox(x, self.step) - self.sol)

    def _fista(self):
        x = self.z - self.step * self.f2.grad(self.z)
        x[:] = self.f1.prox(x, self.step)
        tn = (1. + np.sqrt(1. + 4. * self.t**2.)) / 2.
        self.z[:] = x + (self.t - 1.) / tn * (x - self.sol)
        self.t = tn
        self.sol[:] = x

    def _post(self):
        del self._algo, self.f1, self.f2
        if self.method is 'FISTA':
            del self.z, self.t


class generalized_forward_backward(solver):
    r"""
    Generalized forward-backward proximal splitting algorithm.

    This algorithm solves convex optimization problems composed of the sum of
    any number of non-smooth (or smooth) functions.

    See generic attributes descriptions of the
    :class:`pyunlocbox.solvers.solver` base class.

    Parameters
    ----------
    lambda_ : float, optional
        A relaxation parameter bounded by 0 and 1. Default is 1.

    Notes
    -----
    This algorithm requires each function to either implement the
    :meth:`pyunlocbox.functions.func.prox` method or the
    :meth:`pyunlocbox.functions.func.grad` method.

    See :cite:`raguet2013generalizedFB` for details about the algorithm.

    Examples
    --------
    >>> from pyunlocbox import functions, solvers
    >>> import numpy as np
    >>> y = [0.01, 0.2, 8, 0.3, 0 , 0.03, 7]
    >>> x0 = np.zeros(len(y))
    >>> f1 = functions.norm_l2(y=y)
    >>> f2 = functions.norm_l1()
    >>> solver = solvers.generalized_forward_backward(lambda_=1, step=0.5)
    >>> ret = solvers.solve([f1, f2], x0, solver)
    Solution found after 2 iterations:
        objective function f(sol) = 1.463100e+01
        stopping criterion: RTOL
    >>> ret['sol']
    array([ 0. ,  0. ,  7.5,  0. ,  0. ,  0. ,  6.5])

    """

    def __init__(self, lambda_=1, *args, **kwargs):
        super(generalized_forward_backward, self).__init__(*args, **kwargs)
        self.lambda_ = lambda_

    def _pre(self, functions, x0):

        if self.lambda_ <= 0 or self.lambda_ > 1:
            raise ValueError('Lambda is bounded by 0 and 1.')

        self.f = []  # Smooth functions.
        self.g = []  # Non-smooth functions.
        self.z = []
        for f in functions:
            if 'GRAD' in f.cap(x0):
                self.f.append(f)
            elif 'PROX' in f.cap(x0):
                self.g.append(f)
                self.z.append(np.array(x0, copy=True))
            else:
                raise ValueError('Generalized forward-backward requires each '
                                 'function to implement prox() or grad().')

        if self.verbosity is 'HIGH':
            print('INFO: Generalized forward-backward minimizing {} smooth '
                  'functions and {} non-smooth functions.'.format(len(self.f),
                                                                  len(self.g)))

    def _algo(self):

        # Smooth functions.
        grad = np.zeros(self.sol.shape)
        for f in self.f:
            grad += f.grad(self.sol)

        # Non-smooth functions.
        if not self.g:
            self.sol[:] -= self.step * grad  # Reduces to gradient descent.
        else:
            sol = np.zeros(self.sol.shape)
            for i, g in enumerate(self.g):
                tmp = 2 * self.sol - self.z[i] - self.step * grad
                tmp[:] = g.prox(tmp, self.step * len(self.g))
                self.z[i] += self.lambda_ * (tmp - self.sol)
                sol += 1. * self.z[i] / len(self.g)
            self.sol[:] = sol

    def _post(self):
        del self.f, self.g, self.z


class douglas_rachford(solver):
    r"""
    Douglas-Rachford proximal splitting algorithm.

    This algorithm solves convex optimization problems composed of the sum of
    two non-smooth (or smooth) functions.

    See generic attributes descriptions of the
    :class:`pyunlocbox.solvers.solver` base class.

    Parameters
    ----------
    lambda_ : float, optional
        The update term weight. It should be between 0 and 1. Default is 1.

    Notes
    -----
    This algorithm requires the two functions to implement the
    :meth:`pyunlocbox.functions.func.prox` method.

    See :cite:`combettes2007DR` for details about the algorithm.

    Examples
    --------
    >>> from pyunlocbox import functions, solvers
    >>> import numpy as np
    >>> y = [4, 5, 6, 7]
    >>> x0 = np.zeros(len(y))
    >>> f1 = functions.norm_l2(y=y)
    >>> f2 = functions.dummy()
    >>> solver = solvers.douglas_rachford(lambda_=1, step=1)
    >>> ret = solvers.solve([f1, f2], x0, solver, atol=1e-5)
    Solution found after 8 iterations:
        objective function f(sol) = 2.927052e-06
        stopping criterion: ATOL
    >>> ret['sol']
    array([ 3.99939034,  4.99923792,  5.99908551,  6.99893309])

    """

    def __init__(self, lambda_=1, *args, **kwargs):
        super(douglas_rachford, self).__init__(*args, **kwargs)
        self.lambda_ = lambda_

    def _pre(self, functions, x0):

        if self.lambda_ <= 0 or self.lambda_ > 1:
            raise ValueError('Lambda is bounded by 0 and 1.')

        if len(functions) != 2:
            raise ValueError('Douglas-Rachford requires two convex functions.')

        self.f1 = functions[0]
        self.f2 = functions[1]

        self.z = np.array(x0, copy=True)

    def _algo(self):
        tmp = self.f1.prox(2 * self.sol - self.z, self.step)
        self.z[:] = self.z + self.lambda_ * (tmp - self.sol)
        self.sol[:] = self.f2.prox(self.z, self.step)

    def _post(self):
        del self.f1, self.f2, self.z


class primal_dual(solver):
    r"""
    Parent class of all primal-dual algorithms.

    See generic attributes descriptions of the
    :class:`pyunlocbox.solvers.solver` base class.

    Parameters
    ----------
    L : function or ndarray, optional
        The transformation L that maps from the primal variable space to the
        dual variable space. Default is the identity, :math:`L(x)=x`. If `L` is
        an ``ndarray``, it will be converted to the operator form.
    Lt : function or ndarray, optional
        The adjoint operator. If `Lt` is an ``ndarray``, it will be converted
        to the operator form. If `L` is an ``ndarray``, default is the
        transpose of `L`. If `L` is a function, default is `L`,
        :math:`Lt(x)=L(x)`.
    d0: ndarray, optional
        Initialization of the dual variable.

    """

    def __init__(self, L=None, Lt=None, d0=None, *args, **kwargs):
        super(primal_dual, self).__init__(*args, **kwargs)

        if L is None:
            self.L = lambda x: x
        else:
            if callable(L):
                self.L = L
            else:
                # Transform matrix form to operator form.
                self.L = lambda x: np.dot(L, x)

        if Lt is None:
            if L is None:
                self.Lt = lambda x: x
            elif callable(L):
                self.Lt = L
            else:
                self.Lt = lambda x: np.dot(np.transpose(L), x)
        else:
            if callable(Lt):
                self.Lt = Lt
            else:
                self.Lt = lambda x: np.dot(Lt, x)

        self.d0 = d0

    def _pre(self, functions, x0):
        # Dual variable.
        if self.d0 is None:
            self.dual_sol = self.L(x0)
        else:
            self.dual_sol = self.d0

    def _post(self):
        del self.dual_sol, self.d0


class mlfbf(primal_dual):
    r"""
    Monotone + Lipschitz Forward-Backward-Forward primal-dual algorithm.

    This algorithm solves convex optimization problems with objective of the
    form :math:`f(x) + g(Lx) + h(x)`, where :math:`f` and :math:`g` are proper,
    convex, lower-semicontinuous functions with easy-to-compute proximity
    operators, and :math:`h` has Lipschitz-continuous gradient with constant
    :math:`\beta`.

    See generic attributes descriptions of the
    :class:`pyunlocbox.solvers.primal_dual` base class.

    Notes
    -----
    The order of the functions matters: set :math:`f` first on the list,
    :math:`g` second, and :math:`h` third.

    This algorithm requires the first two functions to implement the
    :meth:`pyunlocbox.functions.func.prox` method, and the third function to
    implement the :meth:`pyunlocbox.functions.func.grad` method.

    The step-size should be in the interval :math:`\left] 0, \frac{1}{\beta +
    \|L\|_{2}}\right[`.

    See :cite:`komodakis2015primaldual`, Algorithm 6, for details.

    Examples
    --------
    >>> from pyunlocbox import functions, solvers
    >>> import numpy as np
    >>> y = np.array([294, 390, 361])
    >>> L = np.array([[5, 9, 3], [7, 8, 5], [4, 4, 9], [0, 1, 7]])
    >>> x0 = np.zeros(len(y))
    >>> f = functions.dummy()
    >>> f._prox = lambda x, T: np.maximum(np.zeros(len(x)), x)
    >>> g = functions.norm_l2(lambda_=0.5)
    >>> h = functions.norm_l2(y=y, lambda_=0.5)
    >>> max_step = 1/(1 + np.linalg.norm(L, 2))
    >>> solver = solvers.mlfbf(L=L, step=max_step/2.)
    >>> ret = solvers.solve([f, g, h], x0, solver, maxit=1000, rtol=0)
    Solution found after 1000 iterations:
        objective function f(sol) = 1.833865e+05
        stopping criterion: MAXIT
    >>> ret['sol']
    array([ 1.,  1.,  1.])

    """

    def _pre(self, functions, x0):
        super(mlfbf, self)._pre(functions, x0)

        if len(functions) != 3:
            raise ValueError('MLFBF requires 3 convex functions.')

        self.f = functions[0]
        self.g = functions[1]
        self.h = functions[2]

    def _algo(self):
        y1 = self.sol - self.step * (self.h.grad(self.sol) +
                                     self.Lt(self.dual_sol))
        y2 = self.dual_sol + self.step * self.L(self.sol)
        p1 = self.f.prox(y1, self.step)
        p2 = _prox_star(self.g, y2, self.step)
        q1 = p1 - self.step * (self.h.grad(p1) + self.Lt(p2))
        q2 = p2 + self.step * self.L(p1)
        self.sol[:] = self.sol - y1 + q1
        self.dual_sol[:] = self.dual_sol - y2 + q2

    def _post(self):
        super(mlfbf, self)._post()
        del self.f, self.g, self.h


class projection_based(primal_dual):
    r"""
    Projection-based primal-dual algorithm.

    This algorithm solves convex optimization problems with objective of the
    form :math:`f(x) + g(Lx)`, where :math:`f` and :math:`g` are proper,
    convex, lower-semicontinuous functions with easy-to-compute proximity
    operators.

    See generic attributes descriptions of the
    :class:`pyunlocbox.solvers.primal_dual` base class.

    Parameters
    ----------
    lambda_ : float, optional
        The update term weight. It should be between 0 and 2. Default is 1.

    Notes
    -----
    The order of the functions matters: set :math:`f` first on the list, and
    :math:`g` second.

    This algorithm requires the two functions to implement the
    :meth:`pyunlocbox.functions.func.prox` method.

    The step-size should be in the interval :math:`\left] 0, \infty \right[`.

    See :cite:`komodakis2015primaldual`, Algorithm 7, for details.

    Examples
    --------
    >>> from pyunlocbox import functions, solvers
    >>> import numpy as np
    >>> y = np.array([294, 390, 361])
    >>> L = np.array([[5, 9, 3], [7, 8, 5], [4, 4, 9], [0, 1, 7]])
    >>> x0 = np.array([500, 1000, -400])
    >>> f = functions.norm_l1(y=y)
    >>> g = functions.norm_l1()
    >>> solver = solvers.projection_based(L=L, step=1.)
    >>> ret = solvers.solve([f, g], x0, solver, maxit=1000, rtol=None, xtol=.1)
    Solution found after 996 iterations:
        objective function f(sol) = 1.045000e+03
        stopping criterion: XTOL
    >>> ret['sol']
    array([0, 0, 0])

    """

    def __init__(self, lambda_=1, *args, **kwargs):
        super(projection_based, self).__init__(*args, **kwargs)
        self.lambda_ = lambda_

    def _pre(self, functions, x0):
        super(projection_based, self)._pre(functions, x0)

        if self.lambda_ <= 0 or self.lambda_ > 2:
            raise ValueError('Lambda is bounded by 0 and 2.')

        if len(functions) != 2:
            raise ValueError('projection_based requires 2 convex functions.')

        self.f = functions[0]
        self.g = functions[1]

    def _algo(self):
        a = self.f.prox(self.sol - self.step *
                        self.Lt(self.dual_sol), self.step)
        l = self.L(self.sol)
        b = self.g.prox(l + self.step * self.dual_sol, self.step)
        s = (self.sol - a) / self.step + self.Lt(l - b) / self.step
        t = b - self.L(a)
        tau = np.sum(s**2) + np.sum(t**2)
        if tau == 0:
            self.sol[:] = a
            self.dual_sol[:] = self.dual_sol + (l - b) / self.step
        else:
            theta = self.lambda_ * (np.sum((self.sol - a)**2) / self.step +
                                    np.sum((l - b)**2) / self.step) / tau
            self.sol[:] = self.sol - theta * s
            self.dual_sol[:] = self.dual_sol - theta * t

    def _post(self):
        super(projection_based, self)._post()
        del self.f, self.g



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Test suite for the pyunlocbox package.
"""

from . import test_functions, test_operators, test_solvers
import unittest
import doctest
import os


def gen_recursive_file(root, ext):
    for root, dirnames, filenames in os.walk(root):
        for name in filenames:
            if name.lower().endswith(ext):
                yield os.path.join(root, name)


def test_docstrings(root, ext):
    files = list(gen_recursive_file(root, ext))
    return doctest.DocFileSuite(*files, module_relative=False)


suites = []
suites.append(test_functions.suite)
suites.append(test_operators.suite)
suites.append(test_solvers.suite)
suites.append(test_docstrings('pyunlocbox', '.py'))
suites.append(test_docstrings('.', '.rst'))
suite = unittest.TestSuite(suites)


def run():  # pragma: no cover
    unittest.TextTestRunner(verbosity=2).run(suite)


if __name__ == '__main__':  # pragma: no cover
    run()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Test suite for the functions module of the pyunlocbox package.
"""

import unittest
import numpy as np
import numpy.testing as nptest
import inspect
from pyunlocbox import functions


class FunctionsTestCase(unittest.TestCase):

    def test_func(self):
        """
        Test the func base class.
        First test that all the methods raise a NotImplemented exception.
        Then assign valid methods and test return values.
        """
        x = 10
        T = 1
        f = functions.func()
        self.assertRaises(NotImplementedError, f.eval, x)
        self.assertRaises(NotImplementedError, f.prox, x, T)
        self.assertRaises(NotImplementedError, f.grad, x)
        f.eval = lambda x: x**2 - 5
        f.grad = lambda x: 2 * x
        f.prox = lambda x, T: x + T

        def assert_equivalent(param1, param2):
            x = [[7, 8, 9], [10, 324, -45], [-7, -.2, 5]]
            funcs = inspect.getmembers(functions, inspect.isclass)
            for f in funcs:
                if f[0] not in ['func', 'norm', 'proj']:
                    f1 = f[1](**param1)
                    f2 = f[1](**param2)
                    self.assertEqual(f1.eval(x), f2.eval(x))
                    nptest.assert_array_equal(f1.prox(x, 3), f2.prox(x, 3))
                    if 'GRAD' in f1.cap(x):
                        nptest.assert_array_equal(f1.grad(x), f2.grad(x))

        # Default parameters. Callable or matrices.
        assert_equivalent({}, {'A': None, 'y': 0, 'At': 1})
        assert_equivalent({'y': 3.2}, {'y': lambda: 3.2})
        assert_equivalent({'A': None}, {'A': np.identity(3)})
        assert_equivalent({'A': None}, {'A': 1})
        assert_equivalent({'A': 6.4}, {'A': lambda x: 6.4 * x})
        A = np.array([[-4, 2, 5], [1, 3, -7], [2, -1, 0]])
        assert_equivalent({'A': A}, {'A': A, 'At': A.T})
        assert_equivalent({'A': lambda x: A.dot(x)}, {'A': A, 'At': A})

    def test_dummy(self):
        """
        Test the dummy derived class.
        All the methods should return 0.
        """
        f = functions.dummy()
        self.assertEqual(f.eval(34), 0)
        nptest.assert_array_equal(f.grad(34), [0])
        nptest.assert_array_equal(f.prox(34, 1), [34])
        x = [34, 2, 1.0, -10.2]
        self.assertEqual(f.eval(x), 0)
        nptest.assert_array_equal(f.grad(x), np.zeros(len(x)))
        nptest.assert_array_equal(f.prox(x, 1), x)

    def test_norm_l2(self):
        """
        Test the norm_l2 derived class.
        We test the three methods : eval, grad and prox.
        First with default class properties, then custom ones.
        """
        f = functions.norm_l2(lambda_=3)
        self.assertEqual(f.eval([10, 0]), 300)
        self.assertEqual(f.eval(np.array([-10, 0])), 300)
        nptest.assert_allclose(f.grad([10, 0]), [60, 0])
        nptest.assert_allclose(f.grad([-10, 0]), [-60, 0])
        self.assertEqual(f.eval([3, 4]), 3 * 5**2)
        self.assertEqual(f.eval(np.array([-3, 4])), 3 * 5**2)
        nptest.assert_allclose(f.grad([3, 4]), [18, 24])
        nptest.assert_allclose(f.grad([3, -4]), [18, -24])
        self.assertEqual(f.prox(0, 1), 0)
        self.assertEqual(f.prox(7, 1. / 6), 3.5)
        f = functions.norm_l2(lambda_=4)
        nptest.assert_allclose(f.prox([7, -22], .125), [3.5, -11])

        f = functions.norm_l2(lambda_=1, A=lambda x: 2 * x, At=lambda x: x / 2,
                              y=[8, 12])
        self.assertEqual(f.eval([4, 6]), 0)
        self.assertEqual(f.eval([5, -2]), 256 + 4)
        nptest.assert_allclose(f.grad([4, 6]), 0)
#        nptest.assert_allclose(f.grad([5, -2]), [8, -64])
        nptest.assert_allclose(f.prox([4, 6], 1), [4, 6])

        f = functions.norm_l2(lambda_=2, y=np.fft.fft([2, 4]) / np.sqrt(2),
                              A=lambda x: np.fft.fft(x) / np.sqrt(x.size),
                              At=lambda x: np.fft.ifft(x) * np.sqrt(x.size))
#        self.assertEqual(f.eval(np.fft.ifft([2, 4])*np.sqrt(2)), 0)
#        self.assertEqual(f.eval([3, 5]), 2*np.sqrt(25+81))
        nptest.assert_allclose(f.grad([2, 4]), 0)
#        nptest.assert_allclose(f.grad([3, 5]), [4*np.sqrt(5), 4*3])
        nptest.assert_allclose(f.prox([2, 4], 1), [2, 4])
        nptest.assert_allclose(f.prox([3, 5], 1), [2.2, 4.2])
        nptest.assert_allclose(f.prox([2.2, 4.2], 1), [2.04, 4.04])
        nptest.assert_allclose(f.prox([2.04, 4.04], 1), [2.008, 4.008])

        # Test prox for non-tight matrices A
        L = np.array([[8, 1, 10], [1, 9, 1], [3, 7, 5], [1, 4, 4]])
        f = functions.norm_l2(A=L, tight=False, y=np.array([1, 2, 3, 4]),
                              w=np.array([1, 1, 0.5, 0.75]))
        nptest.assert_allclose(f.eval([1, 1, 1]), 455.0625)
        nptest.assert_allclose(f.grad([1, 1, 1]),
                               [329.625, 262.500, 430.500], rtol=1e-3)
        nptest.assert_allclose(f.prox([1, 1, 1], 1),
                               [-0.887,  0.252,  0.798], rtol=1e-3)
        nptest.assert_allclose(f.prox([6, 7, 3], 1),
                               [-0.345,  0.298,  0.388], rtol=1e-3)
        nptest.assert_allclose(f.prox([10, 0, -5], 1),
                               [1.103,  0.319,  -0.732], rtol=1e-3)

    def test_soft_thresholding(self):
        """
        Test the soft thresholding helper function.
        """
        x = np.arange(-4, 5, 1)
        # Test integer division for complex method.
        Ts = [2]
        y_gold = [[-2, -1, 0, 0, 0, 0, 0, 1, 2]]
        # Test division by 0 for complex method.
        Ts.append([.4, .3, .2, .1, 0, .1, .2, .3, .4])
        y_gold.append([-3.6, -2.7, -1.8, -.9, 0, .9, 1.8, 2.7, 3.6])
        for k, T in enumerate(Ts):
            for cmplx in [False, True]:
                y_test = functions._soft_threshold(x, T, cmplx)
                nptest.assert_array_equal(y_test, y_gold[k])

    def test_norm_l1(self):
        """
        Test the norm_l1 derived class.
        We test the two methods : eval and prox.
        First with default class properties, then custom ones.
        """
        f = functions.norm_l1(lambda_=3)
        self.assertEqual(f.eval([10, 0]), 30)
        self.assertEqual(f.eval(np.array([-10, 0])), 30)
        self.assertEqual(f.eval([-3, 4]), 21)
        nptest.assert_array_equal(f.prox(np.array([[1, -4], [5, -2]]), 1),
                                  [[0, -1], [2, 0]])

    def test_norm_nuclear(self):
        """
        Test the norm_nuclear derived class.
        We test the two methods : eval and prox.
        First with default class properties, then custom ones.
        """
        f = functions.norm_nuclear(lambda_=3)
        self.assertEqual(f.eval(np.diag([10, 0])), 30)
        self.assertEqual(f.eval(np.diag(np.array([-10, 0]))), 30)
        self.assertEqual(f.eval([[-3]]), 9)
        nptest.assert_allclose(f.prox(np.array([[1, 1], [1, 1]]), 1. / 3),
                               [[.5, .5], [.5, .5]])

    def test_norm_tv(self):
        """
        Test the norm_tv derived class.
        We test the grad, div, eval and prox.
        """
        # Test Matrices initialization
        # test for a 1dim matrice (testing with a 5)
        mat1d = np.arange(5) + 1
        # test for a 2dim matrice (testing with a 2x4)
        mat2d = np.array([[2, 3, 0, 1], [22, 1, 4, 5]])
        # test for a 3 dim matrice (testing with a 2x3x2)
        mat3d = np.arange(1, stop=13).reshape(2, 2, 3).transpose((1, 2, 0))
        # test for a 4dim matrice (2x3x2x2)
        mat4d = np.arange(1, stop=25).reshape(2, 2, 2, 3).transpose((2, 3, 1, 0))
        # test for a 5dim matrice (2x2x3x2x2)
        mat5d = np.arange(1, stop=49).reshape(2, 2, 3, 2, 2).transpose((3, 4, 2, 1, 0))

        # Test for evals
        def test_eval():

            # test with 2d matrices
            # test without weight
            f = functions.norm_tv(dim=1)
            xeval = 30
            nptest.assert_array_equal(xeval, f.eval(mat2d))
            f = functions.norm_tv(dim=2)
            xeval = np.array([56.753641295582440])
            nptest.assert_array_equal(xeval, f.eval(mat2d))

            # test with weights
            f = functions.norm_tv(dim=1, wx=3)
            xeval = np.sum(np.array([60, 6, 12, 12]))
            nptest.assert_array_equal(xeval, f.eval(mat2d))
            f = functions.norm_tv(dim=2, wx=0.5, wy=2)
            xeval = np.array([71.1092])
            nptest.assert_array_equal(xeval, np.around(f.eval(mat2d), decimals=4))

            # test with 3d matrices (2x3x2)
            # test without weight
            f = functions.norm_tv(dim=2)
            sol = np.sum(np.array([11.324555320336760, 11.324555320336760]))
            nptest.assert_array_equal(sol, f.eval(mat3d))
            f = functions.norm_tv(dim=3)
            xeval = np.array(49.762944279683104)
            nptest.assert_array_equal(xeval, f.eval(mat3d))

            # test with weights
            f = functions.norm_tv(dim=2, wx=2, wy=3)
            sol = np.sum(np.array([25.4164, 25.4164]))
            nptest.assert_array_equal(sol, np.around(f.eval(mat3d), decimals=4))

            f = functions.norm_tv(dim=3, wx=2, wy=3, wz=.5)
            xeval = np.array([58.3068])
            nptest.assert_array_equal(xeval, np.around(f.eval(mat3d), decimals=4))

        # Test for prox
        def test_prox():

            # Test with 2d matrices
            # Test without weights
            f = functions.norm_tv(tol=10e-4, dim=1)
            gamma = 30
            sol = np.array([[12.003459453582762, 1.999654054641723,
                             2.000691890716554, 3.000691890716554],
                            [11.996540546417238, 2.000345945358277,
                             1.999308109283446, 2.999308109283446]])
            nptest.assert_array_equal(np.around(sol, decimals=5),
                                      np.around((f.prox(mat2d, gamma)),
                                                decimals=5))

            f = functions.norm_tv(tol=10e-4, dim=2)
            gamma = 1.5
            x2d = np.array([[2, 3, 0, 1], [22, 1, 4, 5], [2, 10, 7, 8]])
            sol = np.array([[3.44427, 2.87332, 2.51662, 2.45336],
                            [18.38207, 3.10251, 4.0028, 4.64074],
                            [4.50809, 6.44118, 6.38421, 6.25082]])
            nptest.assert_array_equal(np.around(sol, decimals=5),
                                      np.around((f.prox(x2d, gamma)),
                                                decimals=5))

            # Test with weights

            # Test with 3d matrices
            # Test without weights
            f = functions.norm_tv(tol=10e-4, dim=2)
            gamma = 42
            sol = np.array([[[3.50087, 9.50087],
                             [3.50000, 9.50000],
                             [3.49913, 9.49913]],
                            [[3.50087, 9.50087],
                             [3.50000, 9.50000],
                             [3.49913, 9.49913]]])
            nptest.assert_array_equal(sol, np.around(f.prox(mat3d, gamma),
                                                     decimals=5))

            f = functions.norm_tv(tol=10e-4, dim=3)
            gamma = 18
            sol = np.array([[[6.5, 6.5], [6.5, 6.5], [6.5, 6.5]],
                            [[6.5, 6.5], [6.5, 6.5], [6.5, 6.5]]])
            nptest.assert_array_equal(sol, np.around(f.prox(mat3d, gamma),
                                                     decimals=1))
            # Test with weights
            f = functions.norm_tv(tol=10e-10, dim=2, wx=5, wy=10)
            gamma = 3
            x3d = np.array([[[1, 10, 19], [2, 11, 20], [3, 12, 21]],
                            [[4, 13, 22], [5, 14, 23], [6, 15, 24]],
                            [[7, 16, 25], [8, 17, 26], [9, 18, 27]]])
            sol = np.array([[[5, 14, 23],
                             [5, 14, 23],
                             [5, 14, 23]],
                            [[5, 14, 23],
                             [5, 14, 23],
                             [5, 14, 23]],
                            [[5, 14, 23],
                             [5, 14, 23],
                             [5, 14, 23]]])
            nptest.assert_array_equal(sol, np.around(f.prox(x3d, gamma)))

            # Test with 4d matrices
            # Test without weights
            f = functions.norm_tv(tol=10e-4, dim=3)
            gamma = 10
            x4d = np.array([[[[1, 28, 55], [10, 37, 64], [19, 46, 73]],
                             [[2, 29, 56], [11, 38, 65], [20, 47, 74]],
                             [[3, 30, 57], [12, 39, 66], [21, 48, 75]]],
                            [[[4, 31, 58], [13, 40, 67], [22, 49, 76]],
                             [[5, 32, 59], [14, 41, 68], [23, 50, 77]],
                             [[6, 33, 60], [15, 42, 69], [24, 51, 78]]],
                            [[[7, 34, 61], [16, 43, 70], [25, 52, 79]],
                             [[8, 35, 62], [17, 44, 71], [26, 53, 80]],
                             [[9, 36, 63], [18, 45, 72], [27, 54, 81]]]])
            sol = np.around(np.array([[[[14, 41, 68], [14, 41, 68], [14, 41, 68]],
                                       [[14, 41, 68], [14, 41, 68], [14, 41, 68]],
                                       [[14, 41, 68], [14, 41, 68], [14, 41, 68]]],
                                      [[[14, 41, 68], [14, 41, 68], [14, 41, 68]],
                                       [[14, 41, 68], [14, 41, 68], [14, 41, 68]],
                                       [[14, 41, 68], [14, 41, 68], [14, 41, 68]]],
                                      [[[14, 41, 68], [14, 41, 68], [14, 41, 68]],
                                       [[14, 41, 68], [14, 41, 68], [14, 41, 68]],
                                       [[14, 41, 68], [14, 41, 68], [14, 41, 68]]]]))
            nptest.assert_array_equal(sol, np.around(f.prox(x4d, gamma)))

            f = functions.norm_tv(tol=10e-4, dim=4)
            gamma = 15
            sol = np.around(np.array([[[[22, 34, 54], [26, 40, 54], [31, 44, 53]],
                                       [[23, 35, 54], [27, 40, 54], [32, 44, 53]],
                                       [[23, 35, 54], [27, 41, 54], [32, 45, 53]]],
                                      [[[24, 36, 54], [28, 41, 54], [32, 45, 53]],
                                       [[24, 36, 54], [28, 42, 53], [33, 45, 53]],
                                       [[24, 37, 54], [29, 42, 53], [33, 46, 53]]],
                                      [[[25, 38, 54], [29, 43, 53], [34, 46, 53]],
                                       [[25, 38, 54], [30, 43, 53], [34, 46, 53]],
                                       [[26, 39, 54], [30, 43, 53], [35, 47, 53]]]]))
            nptest.assert_array_equal(sol, np.around(f.prox(x4d, gamma)))

        # Test with weights
        test_prox()
        test_eval()

    def test_proj_b2(self):
        """
        Test the projection on the L2-ball.
        ISTA and FISTA algorithms for tight and non-tight frames.
        """
        tol = 1e-7

        # Tight frame, radius 0 --> x == y.
        x = np.random.uniform(size=7) + 10
        y = np.random.uniform(size=7) - 10
        f = functions.proj_b2(y=y, epsilon=tol)
        nptest.assert_allclose(f.prox(x, 0), y, atol=tol)

        # Tight frame, random radius --> ||x-y||_2 = radius.
        radius = np.random.uniform()
        f = functions.proj_b2(y=y, epsilon=radius)
        nptest.assert_almost_equal(np.linalg.norm(f.prox(x, 0) - y), radius)

        # Always evaluate to zero.
        self.assertEqual(f.eval(x), 0)

        # Non-tight frame : compare FISTA and ISTA results.
        nx = 30
        ny = 15
        x = np.random.standard_normal(nx)
        y = np.random.standard_normal(ny)
        A = np.random.standard_normal((ny, nx))
        nu = np.linalg.norm(A, ord=2)**2
        f = functions.proj_b2(y=y, A=A, nu=nu, tight=False, method='FISTA',
                              epsilon=5, tol=tol / 10)
        sol_fista = f.prox(x, 0)
        f.method = 'ISTA'
        sol_ista = f.prox(x, 0)
        assert np.allclose(sol_fista, sol_ista, rtol=1e-3)

    def test_independent_problems(self):

        # Parameters.
        N = 3   # independent problems.
        n = 25  # dimensions.

        # Generate some data.
        X = 7 - 10 * np.random.uniform(size=(n, N))
        step = 10 * np.random.uniform()

        # Test all available functions.
        funcs = inspect.getmembers(functions, inspect.isclass)
        for func in funcs:

            # Instanciate the class.
            if func[0] in ['norm_tv']:
                f = func[1](dim=1)  # Each column is one-dimensional.
            else:
                f = func[1]()

            # The combined objective function of the N problems is the sum of
            # each objective.
            if func[0] not in ['func', 'norm', 'norm_nuclear', 'proj']:
                res = 0
                for iN in range(N):
                    res += f.eval(X[:, iN])
                nptest.assert_array_almost_equal(res, f.eval(X))

            # Each column is the prox of one of the N problems.
            # TODO: norm_tv shoud pass this test. Why is there a difference ?
            if func[0] not in ['func', 'norm', 'norm_nuclear', 'norm_tv',
                               'proj']:
                res = np.zeros((n, N))
                for iN in range(N):
                    res[:, iN] = f.prox(X[:, iN], step)
                nptest.assert_array_almost_equal(res, f.prox(X, step))

            # Each column is the gradient of one of the N problems.
            if func[0] not in ['func', 'norm', 'norm_l1', 'norm_nuclear',
                               'norm_tv', 'proj', 'proj_b2']:
                res = np.zeros((n, N))
                for iN in range(N):
                    res[:, iN] = f.grad(X[:, iN])
                nptest.assert_array_almost_equal(res, f.grad(X))


suite = unittest.TestLoader().loadTestsFromTestCase(FunctionsTestCase)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Test suite for the solvers module of the pyunlocbox package.
"""

import unittest
import numpy as np
import numpy.testing as nptest
from pyunlocbox import functions, solvers


class FunctionsTestCase(unittest.TestCase):

    def test_solve(self):
        """
        Test some features of the solving function.
        """
        y = 5 - 10 * np.random.uniform(size=(15, 4))

        def x0(): return np.zeros(y.shape)
        nverb = {'verbosity': 'NONE'}

        # Function verbosity.
        f = functions.dummy()
        self.assertEqual(f.verbosity, 'NONE')
        f.verbosity = 'LOW'
        solvers.solve([f], x0(), **nverb)
        self.assertEqual(f.verbosity, 'LOW')

        # Input parameters.
        self.assertRaises(ValueError, solvers.solve, [f], x0(), verbosity='??')

        # Addition of dummy function.
        self.assertRaises(ValueError, solvers.solve, [], x0(), **nverb)
        solver = solvers.forward_backward()
        solvers.solve([f], x0(), solver, **nverb)
        # self.assertIsInstance(solver.f1, functions.dummy)
        # self.assertIsInstance(solver.f2, functions.dummy)

        # Automatic solver selection.
        f0 = functions.func()
        f0._eval = lambda x: 0
        f0._grad = lambda x: x
        f1 = functions.func()
        f1._eval = lambda x: 0
        f1._grad = lambda x: x
        f1._prox = lambda x, T: x
        f2 = functions.func()
        f2._eval = lambda x: 0
        f2._prox = lambda x, T: x
        self.assertRaises(ValueError, solvers.solve, [f0, f0], x0(), **nverb)
        ret = solvers.solve([f0, f1], x0(), **nverb)
        self.assertEqual(ret['solver'], 'forward_backward')
        ret = solvers.solve([f1, f0], x0(), **nverb)
        self.assertEqual(ret['solver'], 'forward_backward')
        ret = solvers.solve([f1, f2], x0(), **nverb)
        self.assertEqual(ret['solver'], 'forward_backward')
        ret = solvers.solve([f2, f2], x0(), **nverb)
        self.assertEqual(ret['solver'], 'douglas_rachford')
        ret = solvers.solve([f1, f2, f0], x0(), **nverb)
        self.assertEqual(ret['solver'], 'generalized_forward_backward')

        # Stopping criteria.
        f = functions.norm_l2(y=y)
        tol = 1e-6
        r = solvers.solve([f], x0(), None, tol, None, None, None, None, 'NONE')
        self.assertEqual(r['crit'], 'ATOL')
        self.assertLess(np.sum(r['objective'][-1]), tol)
        self.assertEqual(r['niter'], 10)
        tol = 1e-8
        r = solvers.solve([f], x0(), None, None, tol, None, None, None, 'NONE')
        self.assertEqual(r['crit'], 'DTOL')
        err = np.abs(np.sum(r['objective'][-1]) - np.sum(r['objective'][-2]))
        self.assertLess(err, tol)
        self.assertEqual(r['niter'], 14)
        tol = .1
        r = solvers.solve([f], x0(), None, None, None, tol, None, None, 'NONE')
        self.assertEqual(r['crit'], 'RTOL')
        err = np.abs(np.sum(r['objective'][-1]) - np.sum(r['objective'][-2]))
        err /= np.sum(r['objective'][-1])
        self.assertLess(err, tol)
        self.assertEqual(r['niter'], 14)
        tol = 1e-4
        r = solvers.solve([f], x0(), None, None, None, None, tol, None, 'NONE')
        self.assertEqual(r['crit'], 'XTOL')
        r2 = solvers.solve([f], x0(), maxit=r['niter'] - 1, **nverb)
        err = np.linalg.norm(r['sol'] - r2['sol']) / np.sqrt(x0().size)
        self.assertLess(err, tol)
        self.assertEqual(r['niter'], 12)
        nit = 15
        r = solvers.solve([f], x0(), None, None, None, None, None, nit, 'NONE')
        self.assertEqual(r['crit'], 'MAXIT')
        self.assertEqual(r['niter'], nit)

        # Return values.
        f = functions.norm_l2(y=y)
        ret = solvers.solve([f], x0(), **nverb)
        self.assertEqual(len(ret), 6)
        self.assertIsInstance(ret['sol'], np.ndarray)
        self.assertIsInstance(ret['solver'], str)
        self.assertIsInstance(ret['crit'], str)
        self.assertIsInstance(ret['niter'], int)
        self.assertIsInstance(ret['time'], float)
        self.assertIsInstance(ret['objective'], list)

    def test_forward_backward_fista(self):
        """
        Test FISTA algorithm of forward-backward solver with L1-norm, L2-norm
        and dummy functions.
        """
        y = [4, 5, 6, 7]
        solver = solvers.forward_backward(method='FISTA')
        param = {'solver': solver, 'atol': 1e-5, 'verbosity': 'NONE'}

        # L2-norm prox and dummy gradient.
        f1 = functions.norm_l2(y=y)
        f2 = functions.dummy()
        sol = [3.99996922, 4.99996153, 5.99995383, 6.99994614]
        ret = solvers.solve([f1, f2], np.zeros(len(y)), **param)
        nptest.assert_allclose(ret['sol'], sol)
        self.assertEqual(ret['crit'], 'ATOL')
        self.assertEqual(ret['niter'], 10)

        # Dummy prox and L2-norm gradient.
        f1 = functions.dummy()
        f2 = functions.norm_l2(y=y, lambda_=0.6)
        sol = [3.99867319, 4.99834148, 5.99800978, 6.99767808]
        ret = solvers.solve([f1, f2], np.zeros(len(y)), **param)
        nptest.assert_allclose(ret['sol'], sol)
        self.assertEqual(ret['crit'], 'ATOL')
        self.assertEqual(ret['niter'], 10)

        # L2-norm prox and L2-norm gradient.
        f1 = functions.norm_l2(y=y)
        f2 = functions.norm_l2(y=y)
        sol = [3.99904855, 4.99881069, 5.99857282, 6.99833496]
        ret = solvers.solve([f1, f2], np.zeros(len(y)), **param)
        nptest.assert_allclose(ret['sol'], sol)
        self.assertEqual(ret['crit'], 'MAXIT')

        # L1-norm prox and dummy gradient.
        f1 = functions.norm_l1(y=y)
        f2 = functions.dummy()
        sol = y
        ret = solvers.solve([f1, f2], np.zeros(len(y)), **param)
        nptest.assert_allclose(ret['sol'], sol)
        self.assertEqual(ret['crit'], 'ATOL')
        self.assertEqual(ret['niter'], 6)

        # Dummy prox and L1-norm gradient. As L1-norm possesses no gradient,
        # the algorithm exchanges the functions : exact same solution.
        f1 = functions.dummy()
        f2 = functions.norm_l1(y=y)
        sol = y
        ret = solvers.solve([f1, f2], np.zeros(len(y)), **param)
        nptest.assert_allclose(ret['sol'], sol)
        self.assertEqual(ret['crit'], 'ATOL')
        self.assertEqual(ret['niter'], 6)

        # L1-norm prox and L1-norm gradient. L1-norm possesses no gradient.
        f1 = functions.norm_l1(y=y)
        f2 = functions.norm_l1(y=y)
        self.assertRaises(ValueError, solvers.solve,
                          [f1, f2], np.zeros(len(y)), **param)

        # L1-norm prox and L2-norm gradient.
        f1 = functions.norm_l1(y=y, lambda_=1.0)
        f2 = functions.norm_l2(y=y, lambda_=0.8)
        sol = y
        ret = solvers.solve([f1, f2], np.zeros(len(y)), **param)
        nptest.assert_allclose(ret['sol'], sol)
        self.assertEqual(ret['crit'], 'ATOL')
        self.assertEqual(ret['niter'], 4)

    def test_forward_backward_ista(self):
        """
        Test ISTA algorithm of forward-backward solver with L1-norm, L2-norm
        and dummy functions. Test the effect of step and lambda parameters.
        """
        y = [4, 5, 6, 7]
        # Smaller step size and update rate --> slower convergence.
        solver = solvers.forward_backward(method='ISTA', step=.8, lambda_=.5)
        param = {'solver': solver, 'atol': 1e-5, 'verbosity': 'NONE'}

        # L2-norm prox and dummy gradient.
        f1 = functions.norm_l2(y=y)
        f2 = functions.dummy()
        sol = [3.99915094, 4.99893867, 5.9987264, 6.99851414]
        ret = solvers.solve([f1, f2], np.zeros(len(y)), **param)
        nptest.assert_allclose(ret['sol'], sol)
        self.assertEqual(ret['crit'], 'ATOL')
        self.assertEqual(ret['niter'], 23)

        # L1-norm prox and L2-norm gradient.
        f1 = functions.norm_l1(y=y, lambda_=1.0)
        f2 = functions.norm_l2(y=y, lambda_=0.8)
        sol = [3.99999825, 4.9999979, 5.99999756, 6.99999723]
        ret = solvers.solve([f1, f2], np.zeros(len(y)), **param)
        nptest.assert_allclose(ret['sol'], sol)
        self.assertEqual(ret['crit'], 'ATOL')
        self.assertEqual(ret['niter'], 21)

    def test_douglas_rachford(self):
        """
        Test douglas-rachford solver with L1-norm, L2-norm and dummy functions.
        """
        y = [4, 5, 6, 7]
        solver = solvers.douglas_rachford()
        param = {'solver': solver, 'verbosity': 'NONE'}

        # L2-norm prox and dummy prox.
        f1 = functions.norm_l2(y=y)
        f2 = functions.dummy()
        ret = solvers.solve([f1, f2], np.zeros(len(y)), **param)
        nptest.assert_allclose(ret['sol'], y)
        self.assertEqual(ret['crit'], 'RTOL')
        self.assertEqual(ret['niter'], 35)

        # L2-norm prox and L1-norm prox.
        f1 = functions.norm_l2(y=y)
        f2 = functions.norm_l1(y=y)
        ret = solvers.solve([f1, f2], np.zeros(len(y)), **param)
        nptest.assert_allclose(ret['sol'], y)
        self.assertEqual(ret['crit'], 'RTOL')
        self.assertEqual(ret['niter'], 4)

    def test_generalized_forward_backward(self):
        """
        Test the generalized forward-backward algorithm.
        """
        y = [4, 5, 6, 7]
        L = 4  # Gradient of the smooth function is Lipschitz continuous.
        solver = solvers.generalized_forward_backward(step=.9 / L, lambda_=.8)
        params = {'solver': solver, 'verbosity': 'NONE'}

        # Functions.
        f1 = functions.norm_l1(y=y, lambda_=.7)    # Non-smooth.
        f2 = functions.norm_l2(y=y, lambda_=L / 2.)  # Smooth.

        # Solve with 1 smooth and 1 non-smooth.
        ret = solvers.solve([f1, f2], np.zeros(len(y)), **params)
        nptest.assert_allclose(ret['sol'], y)
        self.assertEqual(ret['niter'], 25)

        # Solve with 1 smooth.
        ret = solvers.solve([f1], np.zeros(len(y)), **params)
        nptest.assert_allclose(ret['sol'], y)
        self.assertEqual(ret['niter'], 77)

        # Solve with 1 non-smooth.
        ret = solvers.solve([f2], np.zeros(len(y)), **params)
        nptest.assert_allclose(ret['sol'], y)
        self.assertEqual(ret['niter'], 18)

        # Solve with 1 smooth and 2 non-smooth.
        ret = solvers.solve([f1, f2, f2], np.zeros(len(y)), **params)
        nptest.assert_allclose(ret['sol'], y)
        self.assertEqual(ret['niter'], 26)

        # Solve with 2 smooth and 2 non-smooth.
        ret = solvers.solve([f2, f1, f2, f1], np.zeros(len(y)), **params)
        nptest.assert_allclose(ret['sol'], y)
        self.assertEqual(ret['niter'], 25)

    def test_mlfbf(self):
        """
        Test the MLFBF solver with arbitrarily selected functions.
        """
        x = [1., 1., 1.]
        L = np.array([[5, 9, 3], [7, 8, 5], [4, 4, 9], [0, 1, 7]])
        max_step = 1 / (1 + np.linalg.norm(L, 2))
        solver = solvers.mlfbf(L=L, step=max_step / 2.)
        params = {'solver': solver, 'verbosity': 'NONE'}

        # L2-norm prox and dummy prox.
        f = functions.dummy()
        f._prox = lambda x, T: np.maximum(np.zeros(len(x)), x)
        g = functions.norm_l2(lambda_=0.5)
        h = functions.norm_l2(y=np.array([294, 390, 361]), lambda_=0.5)
        ret = solvers.solve([f, g, h], np.zeros(len(x)),
                            maxit=1000, rtol=0, **params)
        nptest.assert_allclose(ret['sol'], x, rtol=1e-5)

    def test_projection_based(self):
        """
        Test the projection-based solver with arbitrarily selected functions.
        """
        x = [0, 0, 0]
        L = np.array([[5, 9, 3], [7, 8, 5], [4, 4, 9], [0, 1, 7]])
        solver = solvers.projection_based(L=L, step=1.)
        params = {'solver': solver, 'verbosity': 'NONE'}

        # L1-norm prox and dummy prox.
        f = functions.norm_l1(y=np.array([294, 390, 361]))
        g = functions.norm_l1()
        ret = solvers.solve([f, g], np.array([500, 1000, -400]),
                            maxit=1000, rtol=None, xtol=0.1, **params)
        nptest.assert_allclose(ret['sol'], x, rtol=1e-5)

    def test_solver_comparison(self):
        """
        Test that all solvers return the same and correct solution.
        """

        # Convex functions.
        y = [1, 0, 0.1, 8, -6.5, 0.2, 0.004, 0.01]
        sol = [0.75, 0, 0, 7.75, -6.25, 0, 0, 0]
        w1, w2 = .8, .4
        f1 = functions.norm_l2(y=y, lambda_=w1/2.)  # Smooth.
        f2 = functions.norm_l1(lambda_=w2/2.)       # Non-smooth.
        # f3 = functions.proj_b2(epsilon=0.6)         # Non-smooth.

        # Solvers.
        L = w1  # Lipschitz continuous gradient.
        params = {'step': 1./L, 'lambda_': 0.5}
        slvs = []
        slvs.append(solvers.forward_backward(method='ISTA', **params))
        slvs.append(solvers.forward_backward(method='FISTA', **params))
        slvs.append(solvers.douglas_rachford(**params))
        slvs.append(solvers.generalized_forward_backward(**params))
        # slvs.append(solvers.mlfbf(step=.7/L))  # TODO
        slvs.append(solvers.projection_based(step=1e-3))

        # Compare solutions.
        params = {'rtol': 1e-14, 'verbosity': 'NONE', 'maxit': 1e4}
        niters = [26, 2, 61, 26, 1e4]
        for solver, niter in zip(slvs, niters):
            x0 = np.zeros(len(y))
            ret = solvers.solve([f1, f2], x0, solver, **params)
            # These primal-dual solvers are much slower.
            if type(solver) is solvers.projection_based:
                nptest.assert_allclose(ret['sol'], sol, atol=0.2)
            else:
                nptest.assert_allclose(ret['sol'], sol)
            self.assertEqual(ret['niter'], niter)
            self.assertIs(ret['sol'], x0)  # The initial value was modified.


suite = unittest.TestLoader().loadTestsFromTestCase(FunctionsTestCase)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

r"""
This module implements operators functions :

* :meth:`grad` Gradient function for up to 4 dimensions

* :meth:`div` Divergence function for up to 4 dimensions

"""

import numpy as np


def grad(x, dim=2, **kwargs):
    r"""
    Returns the gradient of the array

    Parameters
    ----------
    dim : int
        Dimension of the grad
    wx :  int
    wy :  int
    wz :  int
    wt :  int
        Weights to apply on each axis

    Returns
    -------
    dx, dy, dz, dt : ndarrays
        Gradients following each axes,
        only the necessary ones are returned

    Examples
    --------
    >>> import pyunlocbox
    >>> import numpy as np
    >>> x = np.arange(16).reshape(4, 4)
    >>> dx, dy = pyunlocbox.operators.grad(x)

    """

    axis = 0
    while axis < len(x.shape):
        if axis >= 0:
            try:
                zero_dx = np.zeros((np.append(np.shape(zero_dx),
                                              np.shape(x)[axis])))
            except NameError:
                zero_dx = np.zeros((1))
        if axis >= 1:
            try:
                zero_dy = np.zeros((np.append(np.shape(zero_dy),
                                              np.shape(x)[axis])))
            except NameError:
                zero_dy = np.zeros((np.shape(x)[0], 1))
        if axis >= 2:
            try:
                zero_dz = np.zeros((np.append(np.shape(zero_dz),
                                              np.shape(x)[axis])))
            except NameError:
                zero_dz = np.zeros((np.shape(x)[0], np.shape(x)[1], 1))
        if axis >= 3:
            try:
                zero_dt = np.zeros((np.append(np.shape(zero_dt),
                                              np.shape(x)[axis])))
            except NameError:
                zero_dt = np.zeros((np.shape(x)[0], np.shape(x)[1],
                                    np.shape(x)[2], 1))
        axis += 1

    if dim >= 1:
        dx = np.concatenate((x[1:, ] - x[:-1, ], zero_dx), axis=0)
        try:
            dx *= kwargs["wx"]
        except (KeyError, TypeError):
            pass

    if dim >= 2:
        dy = np.concatenate((x[:, 1:, ] - x[:, :-1, ], zero_dy), axis=1)
        try:
            dy *= kwargs["wy"]
        except (KeyError, TypeError):
            pass

    if dim >= 3:
        dz = np.concatenate((x[:, :, 1:, ] - x[:, :, :-1, ], zero_dz),
                            axis=2)
        try:
            dz *= kwargs["wz"]
        except (KeyError, TypeError):
            pass

    if dim >= 4:
        dt = np.concatenate((x[:, :, :, 1:, ] - x[:, :, :, :-1, ],
                             zero_dt),
                            axis=3)
        try:
            dt *= kwargs["wt"]
        except (KeyError, TypeError):
            pass

    if dim == 1:
        return dx

    elif dim == 2:
        return dx, dy

    elif dim == 3:
        return dx, dy, dz

    elif dim == 4:
        return dx, dy, dz, dt


def grad1d(x, **kwargs):
    return grad(x, dim=1, **kwargs)


def grad2d(x, **kwargs):
    return grad(x, dim=2, **kwargs)


def grad3d(x, **kwargs):
    return grad(x, dim=3, **kwargs)


def grad4d(x, **kwargs):
    return grad(x, dim=4)


def div(*args, **kwargs):
    r"""
    Returns the divergence of the array

    Parameters
    ----------
    dx :  array_like
    dy :  array_like
    dz :  array_like
    dt :  array_like
        Arrays to operate on

    Returns
    -------
    x : array_like
        Divergence vector

    Examples
    --------
    >>> import pyunlocbox
    >>> import numpy as np
    >>> x = np.arange(16).reshape(4, 4)
    >>> dx, dy = pyunlocbox.operators.grad(x)
    >>> divx = pyunlocbox.operators.div(dx, dy)

    """

    if len(args) == 0:
        raise ValueError("Need to input at least one value")

    if len(args) >= 1:
        dx = args[0]
        try:
            dx *= np.conjugate(kwargs["wx"])
        except KeyError:
            pass

        x = np.concatenate((np.expand_dims(dx[0, ], axis=0),
                            dx[1:-1, ] - dx[:-2, ],
                            -np.expand_dims(dx[-2, ], axis=0)),
                           axis=0)

    if len(args) >= 2:
        dy = args[1]
        try:
            dy *= np.conjugate(kwargs["wy"])
        except KeyError:
            pass

        x += np.concatenate((np.expand_dims(dy[:, 0, ], axis=1),
                             dy[:, 1:-1, ] - dy[:, :-2, ],
                             -np.expand_dims(dy[:, -2, ], axis=1)),
                            axis=1)

    if len(args) >= 3:
        dz = args[2]
        try:
            dz *= np.conjugate(kwargs["wz"])
        except KeyError:
            pass

        x += np.concatenate((np.expand_dims(dz[:, :, 0, ], axis=2),
                             dz[:, :, 1:-1, ] - dz[:, :, :-2, ],
                             -np.expand_dims(dz[:, :, -2, ], axis=2)),
                            axis=2)

    if len(args) >= 4:
        dt = args[3]
        try:
            dt *= np.conjugate(kwargs["wt"])
        except KeyError:
            pass

        x += np.concatenate((np.expand_dims(dt[:, :, :, 0, ], axis=3),
                             dt[:, :, :, 1:-1, ] - dt[:, :, :, :-2, ],
                             -np.expand_dims(dt[:, :, :, -2, ], axis=3)),
                            axis=3)
    return x



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from setuptools import setup


setup(
    name='pyunlocbox',
    version='0.4.0',
    description='A convex optimization toolbox using proximal '
                'splitting methods.',
    long_description=open('README.rst').read(),
    author='Michaël Defferrard (EPFL LTS2) and '
           'Nathanaël Perraudin (EPFL LTS2)',
    author_email='michael.defferrard@epfl.ch, nathanael.perraudin@epfl.ch',
    url='https://github.com/epfl-lts2/pyunlocbox',
    packages=['pyunlocbox', 'pyunlocbox.tests'],
    test_suite='pyunlocbox.tests.test_all.suite',
    install_requires=['numpy', 'scipy'],
    license="BSD",
    keywords='convex optimization',
    platforms='any',
    classifiers=[
        'Topic :: Scientific/Engineering :: Mathematics',
        'Environment :: Console',
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Intended Audience :: Education',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: BSD License',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
    ],
)
